/**
  * @author Oscar Sanchez Caadas. 14 Nov 2011
 * 
 */

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketTimeoutException;
import java.security.KeyStore;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.List;

import javax.crypto.SecretKey;
import javax.net.ServerSocketFactory;
import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLServerSocket;

public class Entity extends Thread implements Serializable {

	/**
	 * Entities which relay packets and manage documents that users submit
	 */
	private static final long serialVersionUID = -6782800338944614400L;
	private int port;
	private boolean libre;
	private int idEntity;
	public String nameUser;
	private SecretKey secretKey;
	public PublicKey publicKey;
	private PrivateKey privateKey;
	private String keySym;
	private List<Entity> listaEntities;
	int timesOrigin;
	String messageId;
	User usuario = null;
	InetAddress localAddr;
	int localPort;
	int ePort;
	//Timeout between the source and the origin Entity. 
	int timeout = 900 * 1000; //**Must be modified with the following formula: [Timeout = (Size File  0,02 seg/KB) + (2 * Tunnel lenght)]
	boolean timeExpired = false;
	private long packetSize;

	// It has to be private so that nobody can get, only the Entity
	private PrivateKey getPrivateKey() {
		return privateKey;
	}

	public void setPrivateKey(PrivateKey privateKey) {
		this.privateKey = privateKey;
	}

	public String nameEntity;
	public int order;

	public SSLServerSocket getServer(int port) throws Exception {

		KeyStore kstore = JKS.JKSstart(nameEntity);
		KeyManagerFactory kmf = KeyManagerFactory.getInstance("SunX509");
		kmf.init(kstore, (nameEntity + "_pw").toCharArray());
		SSLContext sslcontext = SSLContext.getInstance("SSLv3");
		sslcontext.init(kmf.getKeyManagers(), null, null);
		ServerSocketFactory ssf = sslcontext.getServerSocketFactory();
		SSLServerSocket serversocket = (SSLServerSocket) ssf
				.createServerSocket(port);
		return serversocket;

	}

	public void run() {
		try {
			ServerSocket socketServidor = new ServerSocket(port); 

			while (true) {

				// Entity socket opens
				socketServidor.setSoTimeout(0);
				// Client is expected
				Socket sslEntity = socketServidor.accept();//**Must be modified to SSLsocket
				SaveFile inf = new SaveFile();

				if (!timeExpired) {
					String servidor = "localhost";

					// There comes a customer
					System.out.println("Entity acepta usuario.");

					InputStream is = sslEntity.getInputStream();
					StatisticsInputStream in = new StatisticsInputStream(is);
					ObjectInputStream oisPA = new ObjectInputStream(in);
					Packet packet = (Packet) oisPA.readObject();

					packetSize = in.getBytesRead();

					PAutomRoutingReceiver paRec = new PAutomRoutingReceiver();
					System.out.println("Entity: " + getIdEntity());
					packet = paRec.sendThroughPEA(packet, getPrivateKey(),
							getListaEntities());

					Calendar calendar = new GregorianCalendar();
					inf.saveSizePacket(calendar.get(Calendar.HOUR_OF_DAY)
							+ " \t " + calendar.get(Calendar.MINUTE) + " \t "
							+ calendar.get(Calendar.SECOND) + " \t "
							+ calendar.get(Calendar.SECOND) + " \t "
							+ packet.getId_Entity() + " \t "
							+ packet.get_role() + " \t " + packetSize
							+ " \t \n");
					System.out.println("Size Packet: " + packetSize);

					if (!(packet.get_len() == 0)) {
						runNextEntity(packet, servidor);

						// We look at whether it is the origin Entity
						if (packet.get_role().equalsIgnoreCase("origen")) {

							// If it is the origin Entity we take messageId before sending the first packet
							messageId = packet.get_messageId();
							// inf.saveInformationEntities(new Date()+ " \t "+
							// packet.getId_Entity()+
							// " \t INF: Timeout origin Entity running... \t \n");
							usuario = (User) oisPA.readObject();

							// inf.saveInformationEntities(new Date() + " \t "+
							// packet.getId_Entity()+
							// " \t INF: Sending File... \t \n");

							// Open a channel to communicate with the user
							ObjectOutputStream oosPA = new ObjectOutputStream(
									sslEntity.getOutputStream());
							
							// We wait to return to Source Entity
							try {
								socketServidor.setSoTimeout(timeout);
								sslEntity = socketServidor.accept();
								socketServidor.setSoTimeout(0);
								// We continue with the chain
								is = sslEntity.getInputStream();
								in = new StatisticsInputStream(is);
								oisPA = new ObjectInputStream(in);
								packet = (Packet) oisPA.readObject();
								// save the packet size
								packetSize = in.getBytesRead();

								paRec = new PAutomRoutingReceiver();
								packet = paRec.sendThroughPEA(packet,
										getPrivateKey(), getListaEntities());
								
								// We save in log the packet size and the role of the Entity
								calendar = new GregorianCalendar();
								inf.saveSizePacket(calendar
										.get(Calendar.HOUR_OF_DAY)
										+ " \t "
										+ calendar.get(Calendar.MINUTE)
										+ " \t "
										+ calendar.get(Calendar.SECOND)
										+ " \t "
										+ calendar.get(Calendar.SECOND)
										+ " \t "
										+ packet.getId_Entity()
										+ " \t "
										+ packet.get_role()
										+ " \t "
										+ packetSize + " \t \n");
								System.out
										.println("Size Packet: " + packetSize);

								// If it is origin Entity again, check the messageId and confirmation is sent to the user
								if (messageId.equalsIgnoreCase(packet
										.get_messageId())
										&& packet.get_EC() == 1) {
									oosPA.writeObject("File successfully received by the destination Entity. \n");
									// inf.saveInformationEntities(new Date()+
									// " \t "+ packet.toString().length()+
									// " \t INF: Archivo recibido \t File successfully received by the destination Entity \t \n");
									runNextEntity(packet, servidor);
								} else {
									oosPA.writeObject("Failed to save the file. Code error: "
											+ packet.get_EC() + " \n");
									// inf.saveInformationEntities(new Date()+
									// " \t "+ packet.toString().length()+
									// " \t INF: Error al guardar \t Failed to save the file \t \n");
									runNextEntity(packet, servidor);
								}
							} catch (SocketTimeoutException s) {
								packet.set_EC((byte) 2);
								oosPA.writeObject("Socket timed out! The file has not been received by the destination Entity. Error code: "
										+ packet.get_EC() + "\n");
								// inf.saveInformationEntities(new Date()+
								// " \t "+ "-"+
								// " \t INF: Socket timed out! \t The file has not been received by the destination Entity. \t \n");
								timeExpired = true;
							} catch (IOException e) {
								e.printStackTrace();
								break;
							}

						}

					}

				} else {
					// inf.saveInformationEntities(new Date()+ " \t "+ "-"+
					// " \t INF: Socket time out! \t Stops forwarding packets because the timeout has expired. \t \n");
					timeExpired = false;
				}

			}
		} catch (Exception e) {
			e.printStackTrace();
		}

	}

	private void runNextEntity(Packet packet, String servidor) {
		// We launched the following Entity
		int nextPort = listaEntities.get(packet.get_idNextEntity() - 1)
				.getPort();
		ObjectOutputStream oos;
		try {
			Socket nextEntity = new Socket(servidor, nextPort);
			oos = new ObjectOutputStream(nextEntity.getOutputStream());
			// inf.saveInformationEntities(new
			// Date()+" \t "+packet.getId_Entity()+" \t INF: Sending next packet... \t \n");
			oos.writeObject(packet.get_nextPacket());
		} catch (IOException e) {
			e.printStackTrace();
		}

	}

	public int getPort() {
		return port;
	}

	public void setPort(int port) {
		this.port = port;
	}

	public boolean isLibre() {
		return libre;
	}

	public void setLibre(boolean libre) {
		this.libre = libre;
	}

	public void setIdEntity(int idEntity) {
		this.idEntity = idEntity;
	}

	public int getIdEntity() {
		return idEntity;
	}

	public void setNameUsuario(String nameUsuario) {
		this.nameUser = nameUsuario;
	}

	public String getNameUsuario() {
		return nameUser;
	}

	public String getNameEntity() {
		return nameEntity;
	}

	public void setNameEntity(String nameEntity) {
		this.nameEntity = nameEntity;
	}

	public PublicKey getPublicKey() {
		return publicKey;
	}

	public void setPublicKey(PublicKey publicKey) {
		this.publicKey = publicKey;
	}

	public int getOrder() {
		return order;
	}

	public void setOrder(int order) {
		this.order = order;
	}

	public SecretKey getSecretKey() {
		return secretKey;
	}

	public void setSecretKey(SecretKey secretKey) {
		this.secretKey = secretKey;
	}

	public String getKeySym() {
		return keySym;
	}

	public void setKeySym(String keySym) {
		this.keySym = keySym;
	}

	public List<Entity> getListaEntities() {
		return listaEntities;
	}

	public void setListaEntities(List<Entity> listaEntities) {
		this.listaEntities = listaEntities;
	}

}
