/**
  * @author Oscar Sanchez Caadas. 14 Nov 2011
 * 
 */

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.UnsupportedEncodingException;
import java.lang.management.ManagementFactory;
import java.net.Socket;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.spec.InvalidKeySpecException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import javax.crypto.SecretKeyFactory;
import javax.crypto.spec.DESKeySpec;

public class PAutomRoutingSender {
	/**
	 *  Anonymous Routing Protocol at the origin Entity
	 */
	String messageId;
	SaveFile log;
	byte[] Padding;

	public void sendThroughPEA(List<Entity> listaEntities, int idEntityDest,
			User usuario, String data, String servidor)
			throws UnsupportedEncodingException, FileNotFoundException {

		// 1. Create a random chain of Entities
		List<Entity> cadena = CreateWay(listaEntities, idEntityDest);

		// 2. Create a symmetric key for each Entity
		CreateKeysSym(cadena);

		// 3. Create package last Entity
		Packet lastPaquet = CreateLastPackage(cadena.get(cadena.size() - 1));

		// 4. Create the package to send Entities with packets of relay, origin and destination Entities
		Packet packet = CreatePackets(lastPaquet, cadena,
				listaEntities.get(idEntityDest - 1), data, usuario);
		log = new SaveFile();
		log.saveLogUser(new
				Date()+" \t "+usuario.getName()+" \t INF: Size packet origin Entity - "+packet.toString().length()+" \t "+packet.getId_Entity()+"\t \n");
		log.saveLogUser(new
				Date()+" \t "+usuario.getName()+" \t INF: Id origin Entity "+packet.getId_Entity()+"\t \n");

		// 5. Sending Packet
		ObjectOutputStream oos;
		try {
			System.out.println(cadena.get(0).getPort());
			Socket sfirstEntity = new Socket(servidor, cadena.get(0).getPort());
			log.saveLogUser(new
					Date()+" \t "+usuario.getName()+" \t INF: Sending packet to origin Entity... - \t \n");
			oos = new ObjectOutputStream(sfirstEntity.getOutputStream());
			oos.writeObject(packet);
			oos.writeObject(usuario);

			ObjectInputStream ois = new ObjectInputStream(
					sfirstEntity.getInputStream());
			String respuesta = (String) ois.readObject();
			System.out.println(respuesta);
			log.saveLogUser(new Date() + " \t " + usuario.getName()
					+ " \t INF: Answer server - " + respuesta + " \t "
					+ packet.getId_Entity() + "\t \n");

		} catch (IOException e) {
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		}
	}

	private Packet CreatePackets(Packet lastPaquet, List<Entity> cadena,
			Entity destEntity, String data, User usuario)
			throws FileNotFoundException {
		// Padding small untill destiny Entity
		Padding = new byte[1 * 1024];
		lastPaquet.set_padding(Padding);
		Entity currentEntity = cadena.get(cadena.size() - 2);
		Entity nextEntity = cadena.get(cadena.size() - 1);
		// We keep the origin Entity
		Entity origEntity = cadena.get(0);
		// Choose the penultimate Entity to incorporate the latest package
		
		// Launched a new package that will include all
		Packet packetAll = new Packet();
		// Put the id of the next Entity in the new package
		packetAll.setId_Entity(currentEntity.getIdEntity());
		packetAll.set_padding(Padding);
		
		int len;
		packetAll.set_lenData(data.getBytes().length);
		if (currentEntity.getIdEntity() == destEntity.getIdEntity()) {
			// Encrypt the package with symmetric key of the destiny Entity, and you return the size of this package
			len = packetAll.cyphEx("destiny", usuario.authz, usuario.op, data
					.toString().length(), data, nextEntity.getIdEntity(),
					lastPaquet, currentEntity.getSecretKey());
		} else if (currentEntity.getIdEntity() == origEntity.getIdEntity()) {
			// Encrypt the package with symmetric key of the origin Entity, and you return the size of this package
			messageId = ManagementFactory.getRuntimeMXBean().getName();
			len = packetAll.cyphEx("origen", messageId,
					nextEntity.getIdEntity(), lastPaquet,
					currentEntity.getSecretKey());
			packetAll.set_padding(Padding);

		} else {
			// Encrypt the package with symmetric key of the relay Entity, and you return the size of this package
			len = packetAll.cyphEx("router", nextEntity.getIdEntity(),
					lastPaquet, currentEntity.getSecretKey());
			packetAll.set_padding(Padding);
		}
		
		// Encrypt the package with public key of the Entities
		packetAll.cyphE_Pbk(currentEntity.getKeySym(), len,
				currentEntity.getPublicKey());

		// Chain saved in log
		 SaveFile log = new SaveFile();
		 log.saveLogCadena(new
				 Date()+" \t "+(cadena.size()-1)+" \t "+cadena.get(cadena.size()-1).getNameEntity()+" \t \n");

		// Create chain
		for (int i = cadena.size() - 2; i > 0; i--) {
			// Save chain in log
			log.saveLogCadena(new Date() + " \t " + i + " \t "+
					cadena.get(i).getNameEntity() + " \t \n");
			System.out.println(nextEntity.getIdEntity());
			// Create next Entity with current Entity as packet
			currentEntity = cadena.get(i - 1);
			nextEntity = cadena.get(i);

			// Create packet of current Entity with previous Entity
			Packet prevPacket = new Packet();
			prevPacket = packetAll;
			packetAll = new Packet();
			// Set id of next Entity in the new package
			packetAll.setId_Entity(currentEntity.getIdEntity());

			packetAll.set_lenData(data.getBytes().length);
			if (currentEntity.getIdEntity() == destEntity.getIdEntity()) {
				// Create the packet encrypted with symmetric key of the destiny Entity, and you return the size of this package
				len = packetAll.cyphEx("destiny", usuario.authz, usuario.op,
						data.toString().length(), data,
						nextEntity.getIdEntity(), prevPacket,
						currentEntity.getSecretKey());
			} else if (currentEntity.getIdEntity() == origEntity.getIdEntity()) {
				// Create the packet encrypted with symmetric key of the origin Entity, and you return the size of this package
				messageId = ManagementFactory.getRuntimeMXBean().getName();
				len = packetAll.cyphEx("origen", messageId,
						nextEntity.getIdEntity(), prevPacket,
						currentEntity.getSecretKey());
				packetAll.set_padding(Padding);
				// packetAll.set_padding(new byte[0]);
			} else {
				// Create the packet encrypted with symmetric key of the relay Entity, and you return the size of this package
				len = packetAll.cyphEx("router", nextEntity.getIdEntity(),
						prevPacket, currentEntity.getSecretKey());
				packetAll.set_padding(Padding);
			}
			// Encrypt the packet encrypted with symmetric key inside thought public key of the destiny Entity
			packetAll.cyphE_Pbk(currentEntity.getKeySym(), len,
					currentEntity.getPublicKey());

		}
		// Set the error code
		packetAll.set_EC((byte) 0);
		// Set the padding to the first packet to send all packets have a similar size
		packetAll.set_padding(Padding);

		return packetAll;
	}

	private Packet CreateLastPackage(Entity lastEntity) {
		// Create last packet
		Packet lastPack = new Packet();

		// The last packet will contain the ID of last Entity, the secret key encrypted by public key and padding
		lastPack.setId_Entity(lastEntity.getIdEntity());
		String keySymlastEntity = lastEntity.getKeySym();

		// Package the symmetric key and his size with public key
		lastPack.cyphE_Pbk(keySymlastEntity, 0, lastEntity.getPublicKey());

		return lastPack;
	}

	private void CreateKeysSym(List<Entity> cadena) {
		try {
			for (int i = 0; i < cadena.size(); i++) {
				PasswordGenerator genPwd = new PasswordGenerator();
				cadena.get(i).setKeySym(genPwd.getPassword());
				SecretKeyFactory skf;
				skf = SecretKeyFactory.getInstance("DES");
				DESKeySpec kspec = new DESKeySpec(cadena.get(i).getKeySym()
						.getBytes("UTF8"));
				cadena.get(i).setSecretKey(skf.generateSecret(kspec));

			}
		} catch (NoSuchAlgorithmException e) {
			e.printStackTrace();
		} catch (InvalidKeyException e) {
			e.printStackTrace();
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
		} catch (InvalidKeySpecException e) {
			e.printStackTrace();
		}

	}

	public List<Entity> CreateWay(List<Entity> listaEntities, int idEntityDest) {

		List<Entity> cadena = new ArrayList<Entity>();
		int order = 0;
		boolean isEntDest = false;
		System.out.println("destiny Entity: " + idEntityDest);
		List<Entity> listaEntitiesAux = new ArrayList<Entity>();
		for (int i = 0; i < listaEntities.size() - 1; i++) {
			listaEntitiesAux.add(i, listaEntities.get(i));
		}
		// Entity randomly chose a origin without choosing again destiny Entity
		int idEntRnd;
		int idEntityOrig;
		do {
			idEntRnd = (int) (Math.random() * listaEntitiesAux.size());
			listaEntitiesAux.get(idEntRnd).setOrder(order++);
			cadena.add(listaEntitiesAux.get(idEntRnd));
			idEntityOrig = listaEntitiesAux.get(idEntRnd).getIdEntity();
		} while (idEntityDest == idEntityOrig);
		System.out.println("origin Entity: " + idEntityOrig);

		// Once we have the origin and destiny Entity, create a random chain where the destiny Entity remain between 
		//the origin Entity and the next time appears the origin Entity
		while (!listaEntitiesAux.isEmpty()) {
			int rndNum = (int) (Math.random() * listaEntitiesAux.size());
			// Add to chain the Entity and remove it from list, if it is not origin cause destiny has to appear before
			if (idEntityOrig != listaEntitiesAux.get(rndNum).getIdEntity()
					|| (isEntDest == true)) {
				listaEntitiesAux.get(rndNum).setOrder(order++);
				cadena.add(listaEntitiesAux.get(rndNum));
				// Look at if it is origin Entity
				if ((idEntityDest == listaEntitiesAux.get(rndNum).getIdEntity())) {
					isEntDest = true;
				}
				listaEntitiesAux.remove(rndNum);

			}

		}
		// Add one more Entity to avoid that origin Entity was last 
		while (cadena.get(cadena.size() - 1).getIdEntity() == idEntityOrig) {
			int rndNum = (int) (Math.random() * listaEntities.size());
			cadena.add(listaEntities.get(rndNum));
			System.out.println("Added one more Entity to chain");
		}
		return cadena;
	}

}
